//
//	Decoder for G.729A CS-ACELP Speech Codec
// 
//		LP_analysis.cpp : Modules for 
//			- Decoding of LP filter parameters [A.4.1.1]			
//			- Same file as in the encoder
//

#include "StdAfx.h"
#include "LP_analysis.h"


/*********************************************************************
/*
/*	  Function to do windowing, autocorrelation [A.3.2.1]
/*				 and Levinson-Durbin algorithm  [A.3.2.2]
/*    Input: s[240], start
/*    Output: s_p[240], a[10]
/*
/*********************************************************************/
int window_and_autocorrelate(float s[240], int start, float s_p[240], float a[10], FILE *fout[15])
{
	float w_lp[240];
	float aa[10][11];
	float E[11], k[11];
	float r[11], r_p[11];
	float w_lag[11];
	float temp;
	int i,j;
	int offset;
	
	// Low Pass Windowing
	offset = (start + 120)%240;
	for (i=0; i<240; i++)
	{
		if (i<200)
			w_lp[i] = 0.54 - 0.46*cos(2*pi*i/399);
		else
			w_lp[i] = cos(2*pi*(i-200)/159);
		s_p[i] = w_lp[i] * s[(offset+i)%240];
	}

	// Compute autocorrelation coefficients
	for (i=0; i<=10; i++)
	{
		r[i] = 0;
		for (j=i; j<240; j++)
			r[i] += s_p[j]*s_p[j-i];
	}
	r[0] = __max(r[0], 1.0);
	r_p[0] = 1.0001 * r[0];
	temp = 2*pi*f0/fs;
	temp = -0.5*temp*temp;
	for (i=1; i<=10; i++)
	{		
		w_lag[i] = exp(temp*i*i);
		r_p[i] = w_lag[i] * r[i];
	}		
	
	// Levinson-Durbin Algorithm
	E[0] = r_p[0];
	for (i=1; i<=10; i++)
	{
		aa[0][i-1] = 1;
		for (j=0,k[i]=0; j<=i-1; j++)
			k[i] += - aa[j][i-1]*r_p[i-j];
		k[i] /= E[i-1];
		aa[i][i] = k[i];
		for (j=1; j<=i-1; j++)
			aa[j][i] = aa[j][i-1] + k[i]*aa[i-j][i-1];
		E[i] = (1 - k[i]*k[i]) * E[i-1];
	}
	for (j=0; j<10; j++)
		a[j] = aa[j+1][10];
	
	for (i=0; i<240; i++)
	{
		fprintf(fout[1],"%f\n",s[(offset+i)%240]);
		fprintf(fout[2],"%f\n",s_p[i]);
	}
	for (i=0; i<=10; i++)
	{
		fprintf(fout[3],"%f\n",r[i]);
		fprintf(fout[4],"%f\n",r_p[i]);
	}
	for (i=0; i<10; i++)	
		fprintf(fout[5],"%f\n",a[i]);

	return 0;
}


/*********************************************************************
/*
/*	  Function to calculate Chebyshev's polynomial [A.3.2.3]
/*    Input: w, f[5]
/*    Output: C(x) = C(cos(w))
/*
/*********************************************************************/
float cheby_poly(float w, float f[5])
{
	float b[6];
	float x;
	int k;
	x = cos(w);
	for (k=3,b[5]=0,b[4]=1; k>=0; k--)	
		b[k] = 2*x*b[k+1] - b[k+2] + f[3-k];
	return(x*b[0] - b[1] + f[4]/2);
}


/************************************************************************
/*
/*	  Function to convert LP coefficients to LSP coefficients [A.3.2.3]
/*    Input: a[10] LP coefficients
/*    Output: q[10] LSP coefficients
/*            w[10] LSF
/*
/************************************************************************/
int LP2LSP(float a[10], float w[10], float q[10], FILE *fout[15])
{
	float C1[50], C2[50];
	float f1[5], f2[5];		
	float wstart, wend, wmid, wtemp;
	float Cstart, Cend, Cmid;
	int i,j,count;	
		

	// F1(z) is defined as 1 + f1(0)z^-1 + f1(1)z^-2 + f1(2)z^-3 + f1(3)z^-4 + f1(4)z^-5 + higher order term in z^-1
	// F2(z) is defined as 1 + f2(0)z^-1 + f2(1)z^-2 + f2(2)z^-3 + f2(3)z^-4 + f2(4)z^-5 + higher order term in z^-1
	f1[0] = a[0] + a[9] - 1;
	f2[0] = a[0] - a[9] + 1;
	for (i=1; i<5; i++)
	{
		f1[i] = a[i] + a[9-i] - f1[i-1];
		f2[i] = a[i] - a[9-i] + f2[i-1];
	}

	for (i=0; i<50; i++)
	{
		wtemp = i * pi / 49;
		C1[i] = cheby_poly(wtemp,f1);
		C2[i] = cheby_poly(wtemp,f2);
	}

	for (i=0, count=0; i<49; i++)
	{
		// Check for the existence of a root for F1(z)
		if (sign(C1[i]) != sign(C1[i+1]))
		{
			wstart = i * pi / 49;
			wend = (i+1) * pi / 49;
			Cstart = C1[i];
			Cend = C1[i+1];
			// The sign interval is divided two times for tracking the root of the polynomial
			for (j=0; j<2; j++)
			{
				wmid = (wstart + wend) / 2;
				Cmid = cheby_poly(wmid,f1);
				if (sign(Cmid) == sign(Cstart))
				{
					Cstart = Cmid;
					wstart = wmid;
				}
				else
				{
					Cend = Cmid;
					wend = wmid;
				}
			}
			if (sign(Cmid) == sign(Cstart))   // The root is between wmid and wend
			{	// The root is estimated by linear interpolating between wmid and wend				
				w[count++] = sign(Cmid)*Cmid/sign(Cend-Cmid)/(Cend - Cmid)*(wend - wmid) + wmid;
			}
			else   //The root is between wmid and wstart
			{	// The root is estimated by linear interpolating between wmid and wstart
				w[count++] = sign(Cstart)*(Cstart)/sign(Cmid-Cstart)/(Cmid - Cstart)*(wmid - wstart) + wstart;
			}

		}

		// Check for the existence of a root for F2(z)
		if (sign(C2[i]) != sign(C2[i+1]))
		{
			wstart = i * pi / 49;
			wend = (i+1) * pi / 49;
			Cstart = C2[i];
			Cend = C2[i+1];
			// The sign interval is divided two times for tracking the root of the polynomial
			for (j=0; j<2; j++)
			{
				wmid = (wstart + wend) / 2;
				Cmid = cheby_poly(wmid,f2);
				if (sign(Cmid) == sign(Cstart))
				{
					Cstart = Cmid;
					wstart = wmid;
				}
				else
				{
					Cend = Cmid;
					wend = wmid;
				}
			}
			if (sign(Cmid) == sign(Cstart))   // The root is between wmid and wend
			{	// The root is estimated by linear interpolating between wmid and wend				
				w[count++] = sign(Cmid)*Cmid/sign(Cend-Cmid)/(Cend - Cmid)*(wend - wmid) + wmid;
			}
			else   //The root is between wmid and wstart
			{	// The root is estimated by linear interpolating between wmid and wstart
				w[count++] = sign(Cstart)*(Cstart)/sign(Cmid-Cstart)/(Cmid - Cstart)*(wmid - wstart) + wstart;
			}
		}
	}
	
	for (i=0; i<10; i++)
		q[i] = cos(w[i]);

	for (i=0; i<10; i++)
		fprintf(fout[6],"%f\n",w[i]);

	return 0;
}


/****************************************************************************
/*
/*	  Function to interpolate LSP coefficients for the 2 subframes [A.3.2.5]
/*    Input: q[2][10] previous and current LSP coefficients
/*           current  current frame pointer
/*    Output: q_sf[2][10] interpolated LSP coefficients for the 2 subframes
/*
/****************************************************************************/
int interpolate_LSP(float q[2][10], int current, float q_sf[2][10])
{
	int i;
	int previous = 1 - current;

	for (i=0; i<10; i++)
	{
		q_sf[0][i] = 0.5*q[previous][i] + 0.5*q[current][i];
		q_sf[1][i] = q[current][i];
	}
	return 0;
}


/*****************************************************************************
/*
/*	  Function to convert LSP coefficients back to LP coefficients [A.3.2.6]
/*    Input: q[10] LSP coefficients
/*    Output: a[10] LP coefficients
/*
/****************************************************************************/
int LSP2LP(float q[10], float a[10])
{
	float f1[2][5];
	float f2[2][5];
	float f1_p[5];
	float f2_p[5];
	int i,j;
	int previous, present;
	

	// Initialize
	for (i=0; i<5; i++)
	{
		f1[0][i] = f1[1][i] = 0;
		f2[0][i] = f2[1][i] = 0;
	}
	previous = 0;
	present = 1;

	// Calculate f1(0), f1(1), f1(2), f1(3), f1(4) from q1,q3,q5,q7,q9 {q[0], q[2], q[4], q[6], q[8]}
	// Calculate f2(0), f2(1), f2(2), f2(3), f2(4) from q2,q4,q6,q8,q10 {q[1], q[3], q[5], q[7], q[9]}
	// F1(z) is defined as 1 + f1(0)z^-1 + f1(1)z^-2 + f1(2)z^-3 + f1(3)z^-4 + f1(4)z^-5 + higher order term in z^-1
	// F2(z) is defined as 1 + f2(0)z^-1 + f2(1)z^-2 + f2(2)z^-3 + f2(3)z^-4 + f2(4)z^-5 + higher order term in z^-1
	for (i=0; i<5; i++)
	{
		present = previous;
		previous = 1 - previous;
		f1[present][0] = f1[previous][0] - 2 * q[i*2];
		f1[present][1] = f1[previous][1] - 2 * q[i*2] * f1[previous][0] + 1;
		f2[present][0] = f2[previous][0] - 2 * q[i*2+1];
		f2[present][1] = f2[previous][1] - 2 * q[i*2+1] * f2[previous][0] + 1;
		for (j=2; j<=__min(i*2+1,4); j++)
		{
			f1[present][j] = f1[previous][j] - 2 * q[i*2] * f1[previous][j-1] + f1[previous][j-2];
			f2[present][j] = f2[previous][j] - 2 * q[i*2+1] * f2[previous][j-1] + f2[previous][j-2];
		}
	}

	for (i=1; i<5; i++)
	{
		f1_p[i] = f1[present][i] + f1[present][i-1];
		f2_p[i] = f2[present][i] - f2[present][i-1];
	}
	f1_p[0] = f1[present][0] + 1;
	f2_p[0] = f2[present][0] - 1;

	for (i=0; i<5; i++)
	{
		a[i] = 0.5*f1_p[i] + 0.5*f2_p[i];
		a[i+5] = 0.5*f1_p[4-i] - 0.5*f2_p[4-i];
	}

	return 0;
}