//
//	Encoder for G.729A CS-ACELP Speech Codec
// 
//		adaptive_codebook_search.cpp : Module for 
//			- [A.3.7] Fixed Codebook
//

#include "StdAfx.h"
#include "adaptive_codebook_search.h"

// Hamming window used for the interpolation
float b30[31] = {
(float) 0.898517,
(float) 0.769271,
(float) 0.448635,
(float) 0.095915,
(float)-0.134333,
(float)-0.178528,
(float)-0.084919,
(float) 0.036952,
(float) 0.095533,
(float) 0.068936,
(float)-0.000000,
(float)-0.050404,
(float)-0.050835,
(float)-0.014169,
(float) 0.023083,
(float) 0.033543,
(float) 0.016774,
(float)-0.007466,
(float)-0.019340,
(float)-0.013755,
(float) 0.000000,
(float) 0.009400,
(float) 0.009029,
(float) 0.002381,
(float)-0.003658,
(float)-0.005027,
(float)-0.002405,
(float) 0.001050,
(float) 0.002780,
(float) 0.002145,
(float) 0.000000
};

/*****************************************************************************
/*
/*	  Function to search the search boundaries for subframe 1 [A.3.7]
/*    Input: T_op Optimal open loop delay
/*    Output: tmin, tmax search boundaries
/*
/****************************************************************************/
void subframe1_search_boundary(int T_op, int *tmin, int *tmax)
{
	*tmin = T_op - 3;
	if (*tmin < 20)
		*tmin = 20;
	*tmax = *tmin + 6;
	if (*tmax > 143)
	{
		*tmax = 143;
		*tmin = *tmax - 6;
	}
}


/*****************************************************************************
/*
/*	  Function to search the search boundaries for subframe 2 [A.3.7]
/*    Input: T1 Optimal open loop delay
/*    Output: tmin, tmax search boundaries
/*
/****************************************************************************/
void subframe2_search_boundary(int T1, int *tmin, int *tmax)
{
	*tmin = T1 - 5;
	if (*tmin < 20)
		*tmin = 20;
	*tmax = *tmin + 9;
	if (*tmax > 143)
	{
		*tmax = 143;
		*tmin = *tmax - 9;
	}
}


/*****************************************************************************
/*
/*	  Function to find out the pitch delay [A.3.7]
/*    Input: tmin, tmax search boundaries
/*           x_b[40] backward filtered target signal
/*           u[200] excitation signal with u[160,...,199] being the current subframe
/*    Output: *T the integer part of the optimum pitch delay
/*			  *frac the fraction part of the optimum pitch delay and takes values -1,0,1
/*
/****************************************************************************/
void pitch_delay(int tmin, int tmax, float x_b[40], float u[200], int *T, int *frac, bool subframe1, float r[40])
{
	int n, k, t;
	int i, j;
	int zero_index = 160;
	float R, Rmax;	
	float u_kt[40];

	for (n=0; n<40; n++)
		u[zero_index+n] = r[n];

	*T = tmin;
	*frac = 0;
	Rmax = 0;

	// Search the integer part within the search boundaries	
	for (n=0; n<40; n++)	
		Rmax += x_b[n]*u[n-tmin+zero_index];	

	for (j=tmin+1; j<=tmax; j++)
	{		
		// Equation A.7
		for (n=0, R=0; n<40; n++)			
			R += x_b[n]*u[n-j+zero_index];

		if (R > Rmax)
		{
			*T = j;
			Rmax = R;
		}
	}
	
	// Check for interpolation conditions
	if ((subframe1 & (*T<85) ) | !subframe1)
	{		
		/* If this is the 1st subframe and the search index < 85, or if this is the 2nd subframe,
		   then the 3 fractional parts T-1/3, T and T+1/3 are interpolated and the maximum
		   correlation term is picked.
			T-1/3: t=2, k=T-1
			T    : t=0, k=T
			T+1/3: t=1, k=T			
		  */
		for (t=0; t<=2; t++)
		{
			k = *T - (t==2);

			// Equation (A.8)
			for (n=0; n<40; n++)
			{
				for (i=0, u_kt[n]=0; i<=9; i++)
				{
					u_kt[n] += u[n-k+i+zero_index]*b30[t+3*i];
					u_kt[n] += u[n-k+1+i+zero_index]*b30[3-t+3*i];
				}
			}
			
			for (n=0, R=0; n<40; n++)
				R += x_b[n]*u_kt[n];
			
			if (t==0)
			{
				*frac = t;
				Rmax = R;
			}			
			else if (R>Rmax)
			{							
				*frac = t;
				Rmax = R;
			}
		}
				
		if (*frac>1)
		{
			*frac = *frac - 3;
			*T = *T + 1;
		}
	}
}		

/*****************************************************************************
/*
/*	  Function to find out the adaptive codebook vector [A.3.7.1]
/*    Input: u[240] excitation signal 
/*           T1, frac1: Optimal integer and fractional part of pitch delay for the 1st subframe
/*			 T2, frac2: Optimal integer and fractional part of pitch delay for the 2nd subframe
/*    Output: v[80] adaptive codebook vector
/*
/****************************************************************************/
void adaptive_codebook_vector(float u[200], int T1, int frac1, float v[40])
{
	int n, i;
	int k, t;
	int zero_index;

	if (frac1 == -1)
	{
		t = 2;
		k = T1 - 1;
	}
	else
	{
		t = frac1;
		k = T1;
	}
	zero_index = 160 - k;
	for (n=0; n<40; n++)
	{
		for (i=0, v[n]=0; i<=9; i++)
			v[n] += u[n+zero_index+i]*b30[t+3*i];
		for (i=0; i<=9; i++)
			v[n] += u[n+zero_index+1+i]*b30[3-t+3*i];
	}

}

/*****************************************************************************
/*
/*	  Function to find out the filtered adaptive codebook vector [A.3.7.3]
/*    Input: v[80] adaptive codebook vector 
/*           h[80] impulse response
/*    Output: y[80] filtered adaptive codebook vector
/*
/****************************************************************************/
void filtered_adaptive_codebook_vector(float v[40], float h[40], float y[40])
{
	int n, i;

	for (n=0; n<40; n++)
		for (i=0, y[n]=0; i<=n; i++)
			y[n] += v[i]*h[n-i];
}

/*****************************************************************************
/*
/*	  Function to find out the dot product of 2 vectors
/*    Input: x[40], y[40]: the 2 vectors whose dot product to be computed
/*    Output: returns the dot product
/*
/****************************************************************************/
float dot_product(float x[40], float y[40])
{
	int n;
	float sum = 0;
	for (n=0; n<40; n++)
		sum += x[n]*y[n];
	return(sum);
}


