//
//	Encoder for G.729A CS-ACELP Speech Codec
// 
//		codebook_search.cpp : Module for 
//			- 
//

#include "StdAfx.h"
#include "codebook_search.h"
#include "adaptive_codebook_search.h"
#include "fixed_codebook_search.h"
#include "quantization_of_gains.h"
#include "open_loop.h"

/*****************************************************************************
/*
/*	  Function to do memory update [A.3.10]
/*    Input: v[40] adaptive codebook vector 
/*           c[40] impulse response
/*    Output: u[200] filtered adaptive codebook vector
/*
/****************************************************************************/
void memory_update(float u[200], float g_p, float v[40], float g_c, float c[40], float x[40], float y[40], float z[40], float ew[10])
{
	int n, i;
	int zero_index = 160;

	for (n=0; n<40; n++)		
		u[zero_index+n] = g_p*v[n] + g_c*c[n];
	
	for (n=0; n<10; n++)
		ew[n] = x[n+30] - g_p*y[n+30] - g_c*z[n+30];

}

/*****************************************************************************
/*
/*	  Function for the main block of the adaptive codebook search
/*    Input: x[80]: target signal
/*			 h[80]: impulse response
/*			 u[240]: excitation buffer
/*			 r[40]: LP residue
/*			 T_op: Optimal open loop delay
/*    Output: *P0: Parity bit fr pitch delay
/*			  *P1: Pitch delay for the 1st subframe
/*			  *P2: Pitch delay for the 2nd subframe
/*			  *g_p1: adaptive codebook gain for the 1st subframe
/*		      *g_p2: adaptive codebook gain for the 2nd subframe
/*			  v[80]: adaptive codebook vector
/*			  y[80]: filtered adaptive codebook vector
/*
/****************************************************************************/
void codebook_search( float a_h[2][10], float x[80], float h[80], float u[240], float r[80], int T_op,
					   bool *P0, int *P1, int *P2, float *g_p1, float *g_p2, float v[80], float y[80], float z[80],
					   int *S1, int *S2, int *C1, int *C2, float c[80], float *g_c1, float *g_c2, float U_h[4], float ew[10])
{
	float x_b[80];
	int tmax1, tmin1, tmax2, tmin2;
	int T1, T2, frac1, frac2;	
	bool subframe1 = true;
	float beta;
	int n;
	float temp;

	response(a_h[0], &r[0], &h[0], &x[0], ew);
	correlation(&x[0], &h[0], &x_b[0]);
	subframe1_search_boundary(T_op, &tmin1, &tmax1);
	pitch_delay(tmin1, tmax1, &x_b[0], &u[0], &T1, &frac1, subframe1, &r[0]);
	adaptive_codebook_vector(&u[0], T1, frac1, &v[0]);
	filtered_adaptive_codebook_vector(&v[0],&h[0],&y[0]);
	temp = dot_product(&y[0],&y[0]);
	if (temp != 0)
		*g_p1 = dot_product(&x[0],&y[0]) / temp;
	else
		*g_p1 = (2*(dot_product(&x[0],&y[0])>0)-1)*GP_BND;

	if (*g_p1 > GP_BND)
		*g_p1 = GP_BND;
	else if (*g_p1 < 0)
		*g_p1 = 0;

	beta = *g_p2;
	fixed_codebook_search_subframe(&x[0], &h[0], &y[0], *g_p1, beta, T1, S1, C1, &c[0]);
	fixed_codebook_gain(&x[0], &y[0], &z[0], &c[0], &h[0], *g_p1, g_c1);
//	gain_prediction(&c[0], g_c1, U_h);
	memory_update(&u[0], *g_p1, &v[0], *g_c1, &c[0], &x[0], &y[0], &z[0], ew);

	response(a_h[1], &r[40], &h[40], &x[40], ew);
	correlation(&x[40], &h[40], &x_b[40]);
	subframe2_search_boundary(T1, &tmin2, &tmax2);	
	pitch_delay(tmin2, tmax2, &x_b[40], &u[40], &T2, &frac2, !subframe1, &r[40]);
	adaptive_codebook_vector(&u[40], T2, frac2, &v[40]);
	filtered_adaptive_codebook_vector(&v[40],&h[40],&y[40]);
	temp = dot_product(&y[40],&y[40]);
	if (temp != 0)
		*g_p2 = dot_product(&x[40],&y[40]) / temp;
	else
		*g_p2 = (2*(dot_product(&x[40],&y[40])>0)-1)*GP_BND;

	if (*g_p2 > GP_BND)
		*g_p2 = GP_BND;
	else if (*g_p2 < 0)
		*g_p2 = 0;

	beta = *g_p1;
	fixed_codebook_search_subframe(&x[40], &h[40], &y[40], *g_p2, beta, T2, S2, C2, &c[40]);
	fixed_codebook_gain(&x[40], &y[40], &z[40], &c[40], &h[40], *g_p2, g_c2);
//	gain_prediction(&c[40], g_c2, U_h);	
	memory_update(&u[40], *g_p2, &v[40], *g_c2, &c[40], &x[40], &y[40], &z[40], ew);


	for (n=0; n<160; n++)
		u[n] = u[n+80];

	if (T1<=85)
		*P1 = 3*(T1-19) + frac1 - 1;
	else
		*P1 = (T1-85) + 197;
	*P2 = 3*(T2 - tmin2) + frac2 + 2;

	*P0 = ((*P1>>2)%2==1) ^ ((*P1>>3)%2==1) ^ ((*P1>>4)%2==1) ^ ((*P1>>5)%2==1) ^ ((*P1>>6)%2==1) ^ ((*P1>>7)%2==1);

}